@@ -141,7 +141,7 @@ public:
 		solver.propagate();
 		solver.assume(index[3].lit);
 		solver.propagate();
-		solver.strategy.restartOnModel = 1;
+		solver.strategies().restartOnModel = true;
 		CPPUNIT_ASSERT(solver.numVars() == solver.numAssignedVars());
 		CPPUNIT_ASSERT_EQUAL(true, e.commitModel(solver));
 		CPPUNIT_ASSERT_EQUAL(true, e.update(solver));
@@ -267,8 +267,8 @@ public:
 	void testTagLiteral() {
 		ModelEnumerator e;
 		SharedContext ctx;
-		Var a = ctx.addVar(Var_t::atom_var);
-		Var b = ctx.addVar(Var_t::atom_var);
+		ctx.addVar(Var_t::atom_var);
+		ctx.addVar(Var_t::atom_var);
 		ctx.startAddConstraints();
 		e.init(ctx);
 		ctx.endInit();
@@ -282,7 +282,7 @@ public:
 	void testIgnoreTagLiteralInPath() {
 		SharedContext ctx;
 		Var a = ctx.addVar(Var_t::atom_var);
-		Var b = ctx.addVar(Var_t::atom_var);
+		ctx.addVar(Var_t::atom_var);
 		Solver& s1 = ctx.startAddConstraints();
 		Solver& s2 = ctx.addSolver();
 		ctx.endInit();
@@ -318,9 +318,9 @@ public:
 		SharedContext ctx;
 		ctx.requestStepVar();
 		Var a = ctx.addVar(Var_t::atom_var);
-		Var b = ctx.addVar(Var_t::atom_var);
+		ctx.addVar(Var_t::atom_var);
 		Solver& s1 = ctx.startAddConstraints();
-		Solver& s2 = ctx.addSolver();
+		ctx.addSolver();
 		ctx.endInit(true);
 		ClauseCreator cc(&s1);
 		cc.start(Constraint_t::learnt_conflict).add(posLit(a)).add(~ctx.stepLiteral()).end();
